
//#define _FILE_OFFSET_BITS 64
#include <stdlib.h>
#include <string.h>
#include "read-frame.h"
#include "VidHeader.h"
#include "zlib.h"

#if defined MSWIN
#define snprintf sprintf_s
#define fseeko _fseeki64
#define ftello _ftelli64
#endif
 

VidHeader :: VidHeader( char * inVideoFileName, FILE * error_outfile  )
{	
	width = 0; height = 0; num_frames = 0;
	pixmap = NULL; compressed_buffer = NULL; uncompressed_buffer = NULL;
	videoFile = NULL;

	this -> error_outfile = error_outfile;
	strncpy( videoFileName, inVideoFileName, sizeof(videoFileName) );

	videoFile = fopen( videoFileName, "rb" );
	if( videoFile == NULL )
	{
		snprintf(errMessage, ERR_BUFFER_LEN, "%s:%d Unable to open following video file in  'rb' mode \r\n %s \r\n", __FILE__, __LINE__, videoFileName);
		displayMessageBox(errMessage, error_outfile); 
		return;
	}

	unsigned char header[ 8 ];

	if( fread( header, 1, 8, videoFile ) < 8 ) {
		snprintf(errMessage, ERR_BUFFER_LEN, "%s:%d Unable to read from video file \n %s \n", __FILE__, __LINE__, videoFileName);
		displayMessageBox(errMessage, error_outfile); 
		return;
	}

	width = header[ 1 ];
	width = ( width << 8 ) | header[ 0 ];

	height = header[ 3 ];
	height = ( height << 8 ) | header[ 2 ];

	num_frames = header[ 7 ];
	num_frames = ( num_frames << 8 ) | header[ 6 ];
	num_frames = ( num_frames << 8 ) | header[ 5 ];
	num_frames = ( num_frames << 8 ) | header[ 4 ];

	current_frame = 0;
	start_frame_number = 0;
	end_frame_number = num_frames;

	uncompressed_buffer_size = ( width * height ) << 1;
	compressed_buffer_size = 2 * uncompressed_buffer_size; //+ ( uncompressed_buffer_size >> 10 ) + 12;
	
	pixmap_size = 3 * width * height;
	pixmap = new unsigned char[pixmap_size];
	compressed_buffer = new unsigned char[compressed_buffer_size];
	uncompressed_buffer = new unsigned char[uncompressed_buffer_size];

	offsets = new off_t[num_frames];
	off_t offset;
	for( int i = 0; i < num_frames; i++ )
	{
		if( fread( header, 1, 8, videoFile ) < 8 ) {
			perror( videoFileName );
			exit(1);
		}

		fseeko( videoFile, ( i + 1 ) << 3, SEEK_SET );

		int j = 7;
		offset = 0;
		while( j >= 0 )
			offset = (offset << 8) | header[ j-- ];	

		offsets[i] = offset;
	}

}

VidHeader::~VidHeader()
{
	fclose(videoFile);
	delete [] pixmap;
	delete [] compressed_buffer;
	delete [] uncompressed_buffer;
}

bool VidHeader:: readNextFrameFromVideoFile()
{
	bool ret_val = readFrameFromVideoFile(current_frame);
	if ( ret_val )
		current_frame++;
		
	return ret_val;
}

bool VidHeader:: readFrameFromVideoFile(int frame_number)
{
	if ( frame_number < 0 || frame_number >= num_frames )
	{
		snprintf(errMessage, ERR_BUFFER_LEN, "%s:%d asked frameNumber %d > num_frames %d\n", __FILE__, __LINE__, frame_number, num_frames);
		displayMessageBox(errMessage, error_outfile);
		return false;
	}

	if( fseeko( videoFile, offsets[frame_number], SEEK_SET ) < 0 )
	{
		snprintf(errMessage, ERR_BUFFER_LEN, "%s:%d cannot seek to offset %Ld for frame_number %d in video_file \n %s\n", __FILE__, __LINE__, offsets	
				[frame_number], frame_number, videoFileName);
		displayMessageBox(errMessage, error_outfile); exit(1);
	}

	int compressed_size;
	if( frame_number <  num_frames - 1 )
		compressed_size = (int) (offsets[frame_number+1] - offsets[frame_number]);
	else
	{
		fseeko( videoFile, 0, SEEK_END );
		off_t end_offset = ftello( videoFile );
		compressed_size = (int) (end_offset - offsets[frame_number]);
		fseeko( videoFile, offsets[frame_number], SEEK_SET );
	}

	memset( pixmap, 0, pixmap_size );
	if ( compressed_size == 0 ) 
		return true;

	if( fread( compressed_buffer, compressed_size, 1, videoFile ) !=  1 ) {
		snprintf(errMessage, ERR_BUFFER_LEN, "%s:%d cannot read frame_number %d in video_file \n %s\n", __FILE__, __LINE__, frame_number, videoFileName);
		displayMessageBox(errMessage, error_outfile); exit(1);
	}

	unsigned long uncompressed_size = uncompressed_buffer_size;
	//uncompress((Bytef *)uncompressed_buffer, (uLongf *)&uncompressed_size, (Bytef *)compressed_buffer, (uLong)compressed_size);
	uncompress(uncompressed_buffer, &uncompressed_size, compressed_buffer, compressed_size);

	if ( uncompressed_size !=  uncompressed_buffer_size )
	{
		snprintf(errMessage, ERR_BUFFER_LEN, "%s:%d error uncompressing frame_number %d in video: %d != %d \n", __FILE__, __LINE__, frame_number, 
								uncompressed_size, uncompressed_buffer_size);
		displayMessageBox(errMessage, error_outfile); exit(1);
	}

	bayer_interpolate_frame( uncompressed_buffer, pixmap, width, height );

	//writePPMimage(frame_number);
	return true;
}

void VidHeader:: writePPMimage(char * imgFileName)
{
	FILE * imgFile;
	imgFile = fopen( imgFileName, "wb" );
	if( imgFile == NULL )
	{
		snprintf(errMessage, ERR_BUFFER_LEN, "%s:%d Unable to open file to write image \r\n %s", __FILE__, __LINE__, imgFileName);
		displayMessageBox(errMessage, error_outfile); return;
	}
	fprintf(imgFile, "P6\n%d %d\n255\n", width, height);
	fwrite(pixmap, width * height * 3, 1, imgFile);
	fclose(imgFile);
}
